# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import uuid
from http import HTTPStatus
from typing import Any
from typing import Dict

import requests
from requests import HTTPError
from requests_toolbelt import user_agent
from requests_toolbelt.adapters import socket_options
from urllib3.util import Retry

from kolena import __name__ as client_name
from kolena import __version__ as client_version
from kolena._utils.endpoints import get_endpoint
from kolena._utils.state import get_client_state
from kolena._utils.state import kolena_initialized
from kolena.errors import IncorrectUsageError
from kolena.errors import NameConflictError
from kolena.errors import NotFoundError
from kolena.errors import RemoteError
from kolena.errors import UnauthenticatedError

__all__ = [
    "get",
    "post",
    "put",
    "delete",
    "raise_for_status",
]

# Give the client 15 seconds to connect to kolena server
# Slightly more than a multiple of 3, as per https://docs.python-requests.org/en/master/user/advanced/#timeouts
CONNECTION_CONNECT_TIMEOUT = 15.05
CONNECTION_READ_TIMEOUT = 60 * 60  # Give kolena server 1 hour to respond to client request

# This only retries for failed DNS lookups, socket connections and connection timeouts.
# HTTPAdapter sets this to 0 by default. https://requests.readthedocs.io/en/latest/_modules/requests/adapters/
# Using the Retry object to configure a backoff which is not supported by using an int here.
MAX_RETRIES = Retry(total=3, connect=3, read=0, redirect=0, status=0, backoff_factor=2)


@kolena_initialized
def _with_default_kwargs(**kwargs: Any) -> Dict[str, Any]:
    client_state = get_client_state()
    default_kwargs = {
        "timeout": (CONNECTION_CONNECT_TIMEOUT, CONNECTION_READ_TIMEOUT),
        "proxies": client_state.proxies,
    }
    default_headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {client_state.jwt_token}",
        "X-Request-ID": uuid.uuid4().hex,
        "User-Agent": user_agent(client_name, client_version),
    }
    return {
        **default_kwargs,
        **kwargs,
        "headers": {**default_headers, **kwargs.get("headers", {})},
    }


@kolena_initialized
def get(endpoint_path: str, params: Any = None, **kwargs: Any) -> requests.Response:
    url = get_endpoint(endpoint_path=endpoint_path)
    with requests.Session() as s:
        s.mount("https://", socket_options.TCPKeepAliveAdapter(max_retries=MAX_RETRIES))
        return s.get(url=url, params=params, **_with_default_kwargs(**kwargs))


@kolena_initialized
def post(endpoint_path: str, data: Any = None, json: Any = None, **kwargs: Any) -> requests.Response:
    url = get_endpoint(endpoint_path=endpoint_path)
    with requests.Session() as s:
        s.mount("https://", socket_options.TCPKeepAliveAdapter(max_retries=MAX_RETRIES))
        return s.post(url=url, data=data, json=json, **_with_default_kwargs(**kwargs))


@kolena_initialized
def put(endpoint_path: str, data: Any = None, json: Any = None, **kwargs: Any) -> requests.Response:
    url = get_endpoint(endpoint_path=endpoint_path)
    with requests.Session() as s:
        s.mount("https://", socket_options.TCPKeepAliveAdapter(max_retries=MAX_RETRIES))
        return s.put(url=url, data=data, json=json, **_with_default_kwargs(**kwargs))


@kolena_initialized
def delete(endpoint_path: str, **kwargs: Any) -> requests.Response:
    url = get_endpoint(endpoint_path=endpoint_path)
    with requests.Session() as s:
        s.mount("https://", socket_options.TCPKeepAliveAdapter(max_retries=MAX_RETRIES))
        return requests.delete(url=url, **_with_default_kwargs(**kwargs))


def raise_for_status(response: requests.Response) -> None:
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        # HTTP 401 is "unauthorized" but used as "unauthenticated"
        raise UnauthenticatedError(response.content)
    if response.status_code == HTTPStatus.NOT_FOUND:
        raise NotFoundError(response.content)
    if response.status_code == HTTPStatus.CONFLICT:
        raise NameConflictError(response.content)
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        raise IncorrectUsageError(response.content)

    try:
        response.raise_for_status()
    except HTTPError:
        raise RemoteError(f"{response.text} ({response.elapsed.total_seconds():0.5f} seconds elapsed)")


@kolena_initialized
def get_connection_args(**kwargs):
    client_state = get_client_state()
    return {"proxies": client_state.proxies}



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List
from typing import Optional

from kolena._api.v1.core import TestCase as API
from kolena._api.v1.workflow import WorkflowType
from kolena.detection import TestCase
from kolena.detection._internal import BaseTestSuite


class TestSuite(BaseTestSuite):
    """
    A test suite groups together one or more test cases.

    For additional functionality, see the associated
    [base class documentation][kolena.detection._internal.test_suite.BaseTestSuite].

    :param name: Name of the test suite to create or load.
    :param version: Optionally specify the version of the test suite to load. When absent, the latest version is loaded.
        Ignored when creating new test suites.
    :param description: Optionally specify a description for a newly created test suite. For existing test suites, this
        description can be edited via [`TestSuite.edit`][kolena.detection._internal.test_suite.BaseTestSuite.edit].
    :param test_cases: Optionally specify a list of test cases to populate a new test suite. For existing test suites,
        test cases can be edited via [`TestSuite.edit`][kolena.detection._internal.test_suite.BaseTestSuite.edit].
    """

    name: str
    """Unique name for this test suite."""

    version: int
    """
    The version of the test suite. This version is automatically incremented each time the suite is edited with
    [`TestSuite.edit`][kolena.detection._internal.test_suite.BaseTestSuite.edit].
    """

    description: str
    """
    Free-form, human-readable description of the test suite. Can be edited at any time via
    [`TestSuite.edit`][kolena.detection._internal.test_suite.BaseTestSuite.edit].
    """

    test_cases: List[TestCase]
    """The test cases within this test suite."""

    _id: int
    _workflow: WorkflowType = WorkflowType.DETECTION

    def __init__(
        self,
        name: str,
        version: Optional[int] = None,
        description: Optional[str] = None,
        test_cases: Optional[List[TestCase]] = None,
        reset: bool = False,
    ):
        super().__init__(name, WorkflowType.DETECTION, version, description, test_cases, reset)

    @classmethod
    def _test_case_from(cls, response: API.EntityData) -> TestCase:
        return TestCase._from(response)



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# flake8: noqa
# autopep8: off
# noreorder
# mypy: ignore-errors
#
# (robots do not like this file)

__name__ = "kolena"
__version__: str


def __version_assign() -> None:
    global __version__
    try:
        from importlib.metadata import version

        __version__ = version(__name__)
    except ModuleNotFoundError:
        import importlib_metadata  # importlib.metadata was introduced to the standard library in 3.8

        __version__ = importlib_metadata.version(__name__)


__version_assign()
del __version_assign


import kolena.errors
import kolena.fr
import kolena.detection
import kolena.classification
from .initialize import initialize

__all__ = [
    "initialize",
    "errors",
    "fr",
    "detection",
    "classification",
    "workflow",
]



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

import pandas as pd
import pandera as pa
from pandera.extensions import register_check_method
from pandera.typing import Series

from kolena._utils.dataframes.validators import validate_df_schema
from kolena._utils.datatypes import LoadableDataFrame
from kolena._utils.serde import as_deserialized_json
from kolena._utils.serde import as_serialized_json
from kolena._utils.serde import with_serialized_columns
from kolena.detection import Inference


@register_check_method()
def _validate_ground_truths(series: pa.typing.Series) -> bool:
    # expectation: Optional[List[dict(data_type=str, data_object=dict)]]
    def validate_cell(cell: List[Dict[str, Any]]) -> bool:
        return all("data_type" in gt and "data_object" in gt and "label" in gt["data_object"] for gt in cell)

    return series.dropna().apply(validate_cell).all()


JSONObject = object


class TestImageDataFrameSchema(pa.SchemaModel):
    locator: Series[pa.typing.String] = pa.Field(coerce=True, _element_wise_validate_image_locator=())
    dataset: Series[pa.typing.String] = pa.Field(coerce=True)
    ground_truths: Series[object] = pa.Field(coerce=True, nullable=True, _validate_ground_truths=())
    metadata: Series[JSONObject] = pa.Field(coerce=True)


class ImageResultDataFrameSchema(pa.SchemaModel):
    test_run_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    model_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    test_sample_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    ignore: Series[pa.typing.Bool] = pa.Field(coerce=True)
    data_type: Series[pa.typing.String] = pa.Field(coerce=True, nullable=True)  # null iff no inferences for sample
    label: Series[pa.typing.String] = pa.Field(coerce=True, nullable=True)  # null iff no inferences for sample
    confidence: Series[pa.typing.Float64] = pa.Field(coerce=True, nullable=True)  # null iff no inferences for sample
    # null if no inferences for sample, or if inference is classification data_type
    polygon: Series[JSONObject] = pa.Field(coerce=True, nullable=True)


class TestImageDataFrame(pa.typing.DataFrame[TestImageDataFrameSchema], LoadableDataFrame["TestImageDataFrame"]):
    def as_serializable(self) -> pd.DataFrame:
        object_columns = ["ground_truths", "metadata"]
        return with_serialized_columns(self, object_columns)

    @classmethod
    def get_schema(cls) -> Type[TestImageDataFrameSchema]:
        return TestImageDataFrameSchema

    @classmethod
    def from_serializable(cls, df: pd.DataFrame) -> "TestImageDataFrame":
        df_deserialized = df.copy()
        df_deserialized["ground_truths"] = df["ground_truths"].apply(as_deserialized_json)
        df_deserialized["metadata"] = df["metadata"].apply(as_deserialized_json)
        return cast(TestImageDataFrame, validate_df_schema(df_deserialized, TestImageDataFrameSchema, trusted=True))


class ImageResultDataFrame(pa.typing.DataFrame[ImageResultDataFrameSchema], LoadableDataFrame["ImageResultDataFrame"]):
    def as_serializable(self) -> pd.DataFrame:
        df_serializable = self.copy()
        df_serializable["polygon"] = df_serializable["polygon"].apply(as_serialized_json)
        return df_serializable

    @classmethod
    def get_schema(cls) -> Type[ImageResultDataFrameSchema]:
        return ImageResultDataFrameSchema

    @classmethod
    def from_image_inference_mapping(
        cls,
        test_run_id: int,
        model_id: int,
        image_inferences: Dict[int, List[Optional[Inference]]],
        ignored_image_ids: List[int],
    ) -> "ImageResultDataFrame":
        records: List[Tuple[int, bool, Optional[str], Optional[str], Optional[float], Optional[dict]]] = []
        for image_id, inferences in image_inferences.items():
            for inference in inferences:
                if inference is None:
                    records.append((image_id, False, None, None, None, None))
                    continue
                inference_dict = inference._to_dict()
                records.append(
                    (
                        image_id,
                        False,
                        inference_dict["data_type"],
                        inference_dict["data_object"]["label"],
                        inference_dict["data_object"]["confidence"],
                        inference_dict["data_object"].get("points", None),
                    ),
                )
        for image_id in ignored_image_ids:
            records.append((image_id, True, None, None, None, None))
        df_stage = pd.DataFrame(
            records,
            columns=["test_sample_id", "ignore", "data_type", "label", "confidence", "polygon"],
        )
        df_stage["test_run_id"] = test_run_id
        df_stage["model_id"] = model_id
        return ImageResultDataFrame(validate_df_schema(df_stage, ImageResultDataFrameSchema, trusted=True))


class LoadInferencesDataFrameSchema(TestImageDataFrameSchema):
    test_case_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    inferences: Series[JSONObject] = pa.Field(coerce=True, nullable=True)  # null if ignored


class LoadInferencesDataFrame(
    pa.typing.DataFrame[LoadInferencesDataFrameSchema],
    LoadableDataFrame["LoadInferencesDataFrame"],
):
    def as_serializable(self) -> pd.DataFrame:
        object_columns = ["ground_truths", "metadata", "inferences"]
        return with_serialized_columns(self, object_columns)

    @classmethod
    def get_schema(cls) -> Type[LoadInferencesDataFrameSchema]:
        return LoadInferencesDataFrameSchema

    @classmethod
    def from_serializable(cls, df: pd.DataFrame) -> "LoadInferencesDataFrame":
        df_deserialized = df.copy()
        df_deserialized["ground_truths"] = df["ground_truths"].apply(as_deserialized_json)
        df_deserialized["metadata"] = df["metadata"].apply(as_deserialized_json)
        df_deserialized["inferences"] = df["inferences"].apply(as_deserialized_json)
        return cast(
            LoadInferencesDataFrame,
            validate_df_schema(df_deserialized, LoadInferencesDataFrameSchema, trusted=True),
        )



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar

from pydantic import validate_arguments

from kolena._utils.frozen import Frozen
from kolena._utils.validators import ValidatorConfig
from kolena.detection._internal.metadata import MetadataElement


TestImageType = TypeVar("TestImageType")


class BaseTestImage(ABC, Frozen):
    locator: str
    dataset: str
    metadata: Dict[str, MetadataElement]

    @validate_arguments(config=ValidatorConfig)
    def __init__(
        self,
        locator: str,
        dataset: Optional[str] = None,
        metadata: Optional[Dict[str, MetadataElement]] = None,
    ):
        self.locator = locator
        self.dataset = dataset or ""
        self.metadata = metadata or {}

    @classmethod
    @abstractmethod
    def _meta_keys(cls) -> List[str]:
        pass

    @classmethod
    @abstractmethod
    def _from_record(cls, record: Any) -> "BaseTestImage":
        pass

    @classmethod
    @abstractmethod
    def _to_record(cls, image: "BaseTestImage") -> Tuple:
        pass



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from pydantic.dataclasses import dataclass

from kolena._api.v1.batched_load import BatchedLoad
from kolena._api.v1.workflow import WorkflowType


class TestImage:
    class Path(str, Enum):
        INIT_LOAD_IMAGES = "/detection/test-sample/load-images/init"

    @dataclass(frozen=True)
    class LoadImagesRequest:
        dataset: Optional[str] = None

    @dataclass(frozen=True)
    class InitLoadImagesRequest(LoadImagesRequest, BatchedLoad.BaseInitDownloadRequest):
        ...


class Model:
    class Path(str, Enum):
        CREATE = "/detection/model/create"
        LOAD_BY_NAME = "/detection/model/load-by-name"
        INIT_LOAD_INFERENCES = "/detection/model/load-inferences/init"
        INIT_LOAD_INFERENCES_BY_TEST_CASE = "/detection/model/load-test-suite-test-cases-inferences/init"
        DELETE = "/detection/model/delete"

    @dataclass(frozen=True)
    class LoadInferencesRequest:
        model_id: int
        batch_size: int
        test_case_id: Optional[int] = None
        test_suite_id: Optional[int] = None

        def __post_init__(self) -> None:
            if (self.test_case_id is None) == (self.test_suite_id is None):
                raise ValueError("must specify exactly one test case or test suite ID")

    @dataclass(frozen=True)
    class InitLoadInferencesRequest(LoadInferencesRequest, BatchedLoad.BaseInitDownloadRequest):
        ...

    @dataclass(frozen=True)
    class LoadInferencesByTestCaseRequest:
        model_id: int
        batch_size: int
        test_suite_id: int

    @dataclass(frozen=True)
    class InitLoadInferencesByTestCaseRequest(
        LoadInferencesByTestCaseRequest,
        BatchedLoad.BaseInitDownloadRequest,
    ):
        ...


class TestCase:
    class Path(str, Enum):
        CREATE = "/detection/test-case/create"
        LOAD_BY_NAME = "/detection/test-case/load-by-name"
        INIT_LOAD_IMAGES = "/detection/test-case/load-images/init"
        COMPLETE_EDIT = "/detection/test-case/edit/complete"


class TestSuite:
    class Path(str, Enum):
        CREATE = "/detection/test-suite/create"
        LOAD_BY_NAME = "/detection/test-suite/load-by-name"
        EDIT = "/detection/test-suite/edit"
        DELETE = "/detection/test-suite/delete"


CustomMetricValue = Union[float, int, None]
CustomMetrics = Dict[str, CustomMetricValue]


class TestRun:
    class Path(str, Enum):
        CREATE_OR_RETRIEVE = "/detection/test-run/create-or-retrieve"
        MARK_CRASHED = "/detection/test-run/mark-crashed"
        INIT_LOAD_REMAINING_IMAGES = "/detection/test-run/load-remaining-images/init"
        UPLOAD_IMAGE_RESULTS = "/detection/test-run/upload-inferences/complete"
        UPLOAD_CUSTOM_METRICS = "/detection/test-run/custom-metrics"

    @dataclass(frozen=True)
    class CreateOrRetrieveRequest:
        model_id: int
        test_suite_ids: List[int]
        config: Optional["Metrics.RunConfig"] = None

    @dataclass(frozen=True)
    class CreateOrRetrieveResponse:
        test_run_id: int

    @dataclass(frozen=True)
    class LoadRemainingImagesRequest:
        test_run_id: int
        batch_size: int

    @dataclass(frozen=True)
    class InitLoadRemainingImagesRequest(LoadRemainingImagesRequest, BatchedLoad.BaseInitDownloadRequest):
        load_all: bool = False

    @dataclass(frozen=True)
    class UploadImageResultsRequest(BatchedLoad.WithLoadUUID):
        test_run_id: int
        reset: bool = False

    @dataclass(frozen=True)
    class UploadImageResultsResponse:
        n_uploaded: int

    @dataclass(frozen=True)
    class UpdateCustomMetricsRequest:
        model_id: int
        metrics: Dict[int, Dict[int, CustomMetrics]]  # testsuite_id -> testcase_id -> CustomMetrics


# work around pydantic nastiness in some environments by declaring this base API object
class _Metrics:
    class RunStrategy(str, Enum):
        F1_OPTIMAL = "F1_OPTIMAL"
        ACCURACY_OPTIMAL = "ACCURACY_OPTIMAL"
        FIXED_GLOBAL_THRESHOLD = "FIXED_GLOBAL_THRESHOLD"


class Metrics(_Metrics):
    @dataclass(frozen=True)
    class RunConfig:
        strategy: _Metrics.RunStrategy
        iou_threshold: float
        params: Optional[Dict[str, Any]] = None

    @dataclass(frozen=True)
    class ComputeRequest:
        test_run_id: int
        workflow_type: WorkflowType
        config: Optional[List["Metrics.RunConfig"]]


# allow referencing e.g. Metrics.RunConfig as type from Metrics.ComputeRequest
Metrics.ComputeRequest.__pydantic_model__.update_forward_refs()  # type: ignore
TestRun.CreateOrRetrieveRequest.__pydantic_model__.update_forward_refs()  # type: ignore



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
from abc import ABCMeta
from abc import abstractmethod
from collections import OrderedDict
from enum import Enum
from typing import Any
from typing import cast
from typing import Dict
from typing import Generic
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

import numpy as np
import pandas as pd
import pandera as pa
from pandera.typing import Series
from pydantic import Extra
from pydantic.dataclasses import dataclass

from kolena._utils.dataframes.validators import validate_df_schema
from kolena._utils.datatypes import get_args
from kolena._utils.datatypes import get_origin
from kolena._utils.datatypes import LoadableDataFrame
from kolena._utils.serde import as_deserialized_json
from kolena._utils.serde import as_serialized_json
from kolena._utils.serde import with_serialized_columns
from kolena._utils.validators import ValidatorConfig

T = TypeVar("T", bound="DataObject")


def _double_under(input: str) -> bool:
    return input.startswith("__") and input.endswith("__")


def _allow_extra(cls: Type[T]) -> bool:
    # `pydantic.dataclasses.is_built_in_dataclass` would have false-positive when a stdlib-dataclass decorated
    # class extends a pydantic dataclass
    return "__pydantic_model__" in vars(cls) and cls.__pydantic_model__.Config.extra == Extra.allow


# used to track data_type string -> TypedDataObject
_DATA_TYPE_MAP = {}


def _get_full_type(obj: "TypedDataObject") -> str:
    data_type = obj._data_type()
    return f"{data_type._data_category()}/{data_type.value}"


def _get_data_type(name: str) -> Optional[Type["TypedDataObject"]]:
    return _DATA_TYPE_MAP.get(name, None)


# used for TypedBaseDataObject to register themselves to be used in dataclass extra fields deserialization
def _register_data_type(cls) -> None:
    full_name = _get_full_type(cls)
    # leverage class inheritance order, only keep base classes of a datatype
    if full_name not in _DATA_TYPE_MAP:
        _DATA_TYPE_MAP[full_name] = cls


def _deserialize_typed_dataobject(value: Dict[Any, Any]) -> Any:
    data_type = _get_data_type(value[DATA_TYPE_FIELD])
    if data_type is None:
        return value

    # 'data_type' is not a real member
    value.pop(DATA_TYPE_FIELD)
    return data_type._from_dict(value)


# best effort to deserialize typed data objects in dataclass extra fields
# note: since a "data_type" string could map to multiple classes through inheritance, only base case would be used.
def _try_deserialize_typed_dataobject(value: Any) -> Any:
    if isinstance(value, list):
        # only attempt deserialization when it is likely this is a list of typed data objects
        if value and isinstance(value[0], dict) and DATA_TYPE_FIELD in value[0]:
            return [_try_deserialize_typed_dataobject(val) for val in value]
    elif isinstance(value, dict):
        if DATA_TYPE_FIELD in value:
            return _deserialize_typed_dataobject(value)

    return value


@dataclass(frozen=True, config=ValidatorConfig)
class DataObject(metaclass=ABCMeta):
    """The base for various objects in `kolena.workflow`."""

    def __str__(self):
        if not _allow_extra(type(self)):
            return self.__repr__()

        # emulate stdlib dataclass _repr_fn implementation, extending extra fields
        fields = [f.name for f in dataclasses.fields(self)]
        extras = [f for f in vars(self) if f not in fields and not _double_under(f)]

        # caveat: dataclass generate `__repr__` for decorated classes, as such calls `__str__` for `DataObject` instead.
        value_str = ", ".join(
            [
                f"{f}={getattr(self, f)}" if isinstance(getattr(self, f), DataObject) else f"{f}={getattr(self, f)!r}"
                for f in fields + extras
            ],
        )
        return f"{self.__class__.__qualname__}({value_str})"

    def _to_dict(self) -> Dict[str, Any]:
        def serialize_value(value: Any) -> Any:
            if isinstance(value, (bool, str, int, float)) or value is None:
                return value
            if isinstance(value, np.generic):  # numpy scalars are common enough to be worth specific handling
                for base_type, numpy_type in [(bool, np.bool_), (int, np.integer), (float, np.inexact)]:
                    if isinstance(value, numpy_type):  # cast if there is a match, otherwise fallthrough
                        return base_type(value)
            if isinstance(value, DataObject):
                return value._to_dict()
            if isinstance(value, (list, tuple, np.ndarray)):
                return [serialize_value(subvalue) for subvalue in value]
            if isinstance(value, dict):
                return {key: serialize_value(subvalue) for key, subvalue in value.items()}
            raise ValueError(f"unsupported value type: '{type(value).__name__}' (value: {value})")

        items = [(field.name, getattr(self, field.name)) for field in dataclasses.fields(type(self))]
        field_names = {field.name for field in dataclasses.fields(type(self))}
        if _allow_extra(type(self)):
            for key, val in vars(self).items():
                if key not in field_names and not _double_under(key):
                    items.append((key, val))
        return OrderedDict([(key, serialize_value(value)) for key, value in items])

    @classmethod
    def _from_dict(cls: Type[T], obj_dict: Dict[str, Any]) -> T:
        def deserialization_invalid_value(field_name: str, field_type: Type, field_value: Any) -> ValueError:
            return ValueError(f"invalid value '{field_value}' provided for field '{field_name}' of type '{field_type}'")

        def default_value_or_raise(field: dataclasses.Field, field_value: Any) -> Any:
            if field.default is not dataclasses.MISSING:
                return field.default
            if field.default_factory is not dataclasses.MISSING:
                return field.default_factory()
            raise deserialization_invalid_value(field.name, field.type, field_value)

        def deserialize_field(
            field: dataclasses.Field,
            field_value: Any,
            field_type: Optional[Type] = None,  # override field.type, for recursively deserializing e.g. List[T]
            attempt_cast: bool = True,
        ) -> Any:
            """
            Custom deserialization following rules applied in ``validate_data_object_type``. This is maintained as a
            faster alternative to deserialize an arbitrarily nested value into a dataclass.
            """
            field_type = field_type or field.type
            origin = get_origin(field_type)  # non-None for typing.X types

            if origin is list:
                (arg,) = get_args(field_type)
                if not isinstance(field_value, list):
                    return default_value_or_raise(field, field_value)
                return [deserialize_field(field, v, field_type=arg) for v in field_value]

            elif origin is dict or origin is Dict:
                (arg_key, arg_val) = get_args(field_type)
                if not isinstance(field_value, dict):
                    return default_value_or_raise(field, field_value)
                entries = [
                    (deserialize_field(field, k, field_type=arg_key), deserialize_field(field, v, field_type=arg_val))
                    for k, v in field_value.items()
                ]
                return OrderedDict(entries)

            elif origin is Union:  # includes Optional[T]
                args = get_args(field_type)
                # first attempt to find a direct match without casting, then if that fails try each type with casting
                for should_cast in [False, True]:
                    for arg in args:  # find the first type arg in the list for which deserialization succeeds
                        try:
                            return deserialize_field(field, field_value, field_type=arg, attempt_cast=should_cast)
                        except (ValueError, AttributeError):  # potential AttributeError when field_value=None
                            continue
                raise deserialization_invalid_value(field.name, field_type, field_value)

            elif origin is tuple or origin is Tuple:
                args = get_args(field_type)
                if not isinstance(field_value, (list, tuple)):  # should really only be list as the input is JSON
                    raise deserialization_invalid_value(field.name, field_type, field_value)
                return tuple(deserialize_field(field, field_value[i], field_type=arg) for i, arg in enumerate(args))

            elif origin is not None:  # unsupported generic
                raise ValueError(f"unsupported type '{field_type}' for field '{field.name}' and value '{field_value}'")

            # if we've reached here, it's not a typing generic, and issubclass is safe to use
            if field_value is None and not issubclass(field_type, type(None)):
                return default_value_or_raise(field, field_value)

            if isinstance(field_value, field_type):
                return field_value

            elif issubclass(field_type, DataObject):
                return field_type._from_dict(field_value)

            if attempt_cast and not issubclass(field_type, type(None)):
                return field_type(field_value)

            raise ValueError(
                f"cast=False, not casting field '{field.name}' to type '{field_type}' with value '{field_value}'",
            )

        items = {f.name: deserialize_field(f, obj_dict.get(f.name, None)) for f in dataclasses.fields(cls) if f.init}
        field_names = {f.name for f in dataclasses.fields(cls)}
        if _allow_extra(cls):
            for key, val in obj_dict.items():
                if key not in field_names:
                    items[key] = _try_deserialize_typed_dataobject(val)
        return cls(**items)

    # integrate with pandas json deserialization
    # https://pandas.pydata.org/docs/user_guide/io.html#fallback-behavior
    def toDict(self) -> Dict:
        return self._to_dict()


class DataType(str, Enum):
    @staticmethod
    def _data_category() -> str:
        raise NotImplementedError


DATA_TYPE_FIELD = "data_type"
U = TypeVar("U", bound=DataType)


@dataclass(frozen=True, config=ValidatorConfig)
class TypedDataObject(Generic[U], DataObject, metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def _data_type() -> U:
        raise NotImplementedError

    def _to_dict(self) -> Dict[str, Any]:
        self_dict = super()._to_dict()
        self_type = self._data_type()
        self_dict[DATA_TYPE_FIELD] = f"{self_type._data_category()}/{self_type.value}"
        return self_dict


_SCALAR_TYPES = [str, bool, int, float]
JSONObject = object


class TestSampleDataFrameSchema(pa.SchemaModel):
    """General-purpose frame used for test samples in isolation or paired with ground truths and/or inferences."""

    test_sample: Series[JSONObject] = pa.Field(coerce=True)
    test_sample_metadata: Optional[Series[JSONObject]] = pa.Field(coerce=True)
    ground_truth: Optional[Series[JSONObject]] = pa.Field(coerce=True, nullable=True)
    inference: Optional[Series[JSONObject]] = pa.Field(coerce=True)


class TestSampleDataFrame(
    pa.typing.DataFrame[TestSampleDataFrameSchema],
    LoadableDataFrame["TestSampleDataFrame"],
):
    @classmethod
    def get_schema(cls) -> Type[TestSampleDataFrameSchema]:
        return TestSampleDataFrameSchema

    def as_serializable(self) -> pd.DataFrame:
        return TestSampleDataFrame._serde(self, True)

    @classmethod
    def from_serializable(cls, df: pd.DataFrame) -> "TestSampleDataFrame":
        df_deserialized = TestSampleDataFrame._serde(df, False)
        df_validated = validate_df_schema(df_deserialized, TestSampleDataFrameSchema, trusted=True)
        return cast(TestSampleDataFrame, df_validated)

    @staticmethod
    def _serde(df: pd.DataFrame, serialize: bool) -> pd.DataFrame:
        serde_function = as_serialized_json if serialize else as_deserialized_json
        df_out = df.copy()
        df_out["test_sample"] = df_out["test_sample"].apply(serde_function)
        if "test_sample_metadata" in df.columns:
            df_out["test_sample_metadata"] = df_out["test_sample_metadata"].apply(serde_function)
        if "ground_truth" in df.columns:
            df_out["ground_truth"] = df_out["ground_truth"].apply(serde_function)
        if "inference" in df.columns:
            df_out["inference"] = df_out["inference"].apply(serde_function)
        return df_out


class TestSuiteTestSamplesDataFrameSchema(pa.SchemaModel):
    """Data frame used for loading test samples grouped by test case."""

    test_case_id: Optional[Series[pa.typing.Int64]] = pa.Field(coerce=True)
    test_sample: Series[JSONObject] = pa.Field(coerce=True)
    test_sample_metadata: Series[JSONObject] = pa.Field(coerce=True)


class TestSuiteTestSamplesDataFrame(
    pa.typing.DataFrame[TestSuiteTestSamplesDataFrameSchema],
    LoadableDataFrame["TestSuiteTestSamplesDataFrame"],
):
    @classmethod
    def get_schema(cls) -> Type[TestSuiteTestSamplesDataFrameSchema]:
        return TestSuiteTestSamplesDataFrameSchema

    def as_serializable(self) -> pd.DataFrame:
        return TestSuiteTestSamplesDataFrame._serde(self, True)

    @classmethod
    def from_serializable(cls, df: pd.DataFrame) -> "TestSuiteTestSamplesDataFrame":
        df_deserialized = TestSuiteTestSamplesDataFrame._serde(df, False)
        df_validated = validate_df_schema(df_deserialized, TestSuiteTestSamplesDataFrameSchema, trusted=True)
        return cast(TestSuiteTestSamplesDataFrame, df_validated)

    @staticmethod
    def _serde(df: pd.DataFrame, serialize: bool) -> pd.DataFrame:
        serde_function = as_serialized_json if serialize else as_deserialized_json
        df_out = df.copy()
        df_out["test_sample"] = df_out["test_sample"].apply(serde_function)
        df_out["test_sample_metadata"] = df_out["test_sample_metadata"].apply(serde_function)
        return df_out


class TestCaseEditorDataFrameSchema(pa.SchemaModel):
    test_case_name: Series[pa.typing.String] = pa.Field(nullable=True)
    test_sample_type: Series[pa.typing.String] = pa.Field(coerce=True)
    test_sample: Series[JSONObject] = pa.Field(coerce=True)  # TODO: validators?
    test_sample_metadata: Series[JSONObject] = pa.Field(coerce=True)
    ground_truth: Series[JSONObject] = pa.Field(coerce=True, nullable=True)
    remove: Series[pa.typing.Bool] = pa.Field(coerce=True)


class TestCaseEditorDataFrame(
    pa.typing.DataFrame[TestCaseEditorDataFrameSchema],
    LoadableDataFrame["TestCaseEditorDataFrame"],
):
    def as_serializable(self) -> pd.DataFrame:
        object_columns = ["test_sample", "test_sample_metadata", "ground_truth"]
        return with_serialized_columns(self, object_columns)

    @classmethod
    def get_schema(cls) -> Type[TestCaseEditorDataFrameSchema]:
        return TestCaseEditorDataFrameSchema

    @classmethod
    def from_serializable(cls, df: pd.DataFrame) -> "TestCaseEditorDataFrame":
        df_deserialized = df.copy()
        df_deserialized["test_sample"] = df["test_sample"].apply(as_deserialized_json)
        df_deserialized["test_sample_metadata"] = df["test_sample_metadata"].apply(as_deserialized_json)
        df_deserialized["ground_truths"] = df["ground_truths"].apply(as_deserialized_json)
        df_validated = validate_df_schema(df_deserialized, TestCaseEditorDataFrameSchema, trusted=True)
        return cast(TestCaseEditorDataFrame, df_validated)


class MetricsDataFrameSchema(pa.SchemaModel):
    test_sample: Optional[Series[JSONObject]] = pa.Field(coerce=True)
    test_case_id: Optional[Series[pa.typing.Int64]] = pa.Field(coerce=True)
    configuration_display_name: Optional[Series[pa.typing.String]] = pa.Field(coerce=True, nullable=True)
    metrics: Series[JSONObject] = pa.Field(coerce=True)


class MetricsDataFrame(pa.typing.DataFrame[MetricsDataFrameSchema], LoadableDataFrame["MetricsDataFrame"]):
    @classmethod
    def get_schema(cls) -> Type[MetricsDataFrameSchema]:
        return MetricsDataFrameSchema

    def as_serializable(self) -> pd.DataFrame:
        return MetricsDataFrame._serde(self, True)

    @classmethod
    def from_serializable(cls, df: pd.DataFrame) -> "MetricsDataFrame":
        df_deserialized = MetricsDataFrame._serde(df, False)
        df_validated = validate_df_schema(df_deserialized, MetricsDataFrameSchema, trusted=True)
        return cast(MetricsDataFrame, df_validated)

    @staticmethod
    def _serde(df: pd.DataFrame, serialize: bool) -> pd.DataFrame:
        serde_function = as_serialized_json if serialize else as_deserialized_json
        df_out = df.copy()
        if "test_sample" in df.columns:
            df_out["test_sample"] = df_out["test_sample"].apply(serde_function)
        df_out["metrics"] = df_out["metrics"].apply(serde_function)
        return df_out



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from typing import Any
from typing import Union

import pandas as pd

from kolena.workflow import DataObject
from kolena.workflow._datatypes import _DATA_TYPE_MAP
from kolena.workflow._datatypes import DATA_TYPE_FIELD


def _serialize_dataobject(x: Any) -> Any:
    if isinstance(x, list):
        return [item._to_dict() if isinstance(item, DataObject) else item for item in x]

    return x._to_dict() if isinstance(x, DataObject) else x


def _deserialize_dataobject(x: Any) -> Any:
    if isinstance(x, list):
        return [_deserialize_dataobject(item) for item in x]

    if isinstance(x, dict) and DATA_TYPE_FIELD in x:
        data = {**x}
        data_type = data.pop(DATA_TYPE_FIELD)
        typed_dataobject = _DATA_TYPE_MAP.get(data_type, None)
        if typed_dataobject:
            return typed_dataobject._from_dict(data)

    return x


def _serialize_json(x: Any) -> Any:
    if isinstance(x, list) or isinstance(x, dict):
        return json.dumps(x)

    return x


def _deserialize_json(x: Any) -> Any:
    if isinstance(x, str):
        try:
            return json.loads(x)
        except Exception:
            ...

    return x


def dataframe_to_csv(df: pd.DataFrame, *args, **kwargs) -> Union[str, None]:
    """
    Helper function to export pandas DataFrame containing annotation or asset to CSV format.

    :param args: positional arguments to `pandas.DataFrame.to_csv`.
    :param kwargs: keyword arguments to `pandas.DataFrame.to_csv`.
    :return: None or str.
    """
    columns = list(df.select_dtypes(include="object").columns)
    df_post = df.select_dtypes(exclude="object")
    df_post[columns] = df[columns].applymap(_serialize_dataobject)
    df_post[columns] = df_post[columns].applymap(_serialize_json)
    return df_post.to_csv(*args, **kwargs)


def dataframe_from_csv(*args, **kwargs) -> pd.DataFrame:
    """
    Helper function to load pandas DataFrame exported to CSV with `dataframe_to_csv`.

    :param args: positional arguments to `pandas.DataFrame.read_csv`.
    :param kwargs: keyword arguments to `pandas.DataFrame.read_csv`.
    :return: DataFrame.
    """
    df = pd.read_csv(*args, **kwargs)
    columns = list(df.select_dtypes(include="object").columns)
    df_post = df.select_dtypes(exclude="object")
    df_post[columns] = df[columns].applymap(_deserialize_json)
    df_post[columns] = df_post[columns].applymap(_deserialize_dataobject)

    return df_post


def dataframe_from_json(*args, **kwargs) -> pd.DataFrame:
    """
    Helper function to load pandas DataFrame containing annotation or asset from JSON file or string.

    :param args: positional arguments to `pandas.DataFrame.read_json`.
    :param kwargs: keyword arguments to `pandas.DataFrame.read_json`.
    :return: DataFrame.
    """
    df = pd.read_json(*args, **kwargs)
    columns = list(df.select_dtypes(include="object").columns)
    df_post = df.select_dtypes(exclude="object")
    df_post[columns] = df[columns].applymap(_deserialize_dataobject)

    return df_post



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
from abc import ABC
from collections import OrderedDict
from contextlib import contextmanager
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Type

import pandas as pd
from pydantic import validate_arguments

from kolena._api.v1.core import TestCase as CoreAPI
from kolena._api.v1.detection import TestCase as API
from kolena._api.v1.workflow import WorkflowType
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.batched_load import DFType
from kolena._utils.batched_load import init_upload
from kolena._utils.batched_load import upload_data_frame
from kolena._utils.consts import BatchSize
from kolena._utils.consts import FieldName
from kolena._utils.dataframes.validators import validate_df_schema
from kolena._utils.frozen import Frozen
from kolena._utils.instrumentation import WithTelemetry
from kolena._utils.serde import from_dict
from kolena._utils.validators import validate_name
from kolena._utils.validators import ValidatorConfig
from kolena.detection._internal import BaseTestImage
from kolena.errors import NotFoundError
from kolena.workflow._validators import assert_workflows_match


class BaseTestCase(ABC, Frozen, WithTelemetry):
    """
    A test case holds a set of images to compute performance metrics against.

    :param name: Name of the test case to create or load.
    :param version: Optionally specify the version of the test case to load. When absent, the latest version is loaded.
        Ignored when creating new test cases.
    :param description: Optionally specify a description for a newly created test case. For existing test cases, this
        description can be edited via [`TestCase.edit`][kolena.detection._internal.BaseTestCase.edit].
    :param images: Optionally provide a list of images and associated ground truths to populate a new test case. For
        existing test cases, images can be edited via [`TestCase.edit`][kolena.detection._internal.BaseTestCase.edit].
    """

    _TestImageClass: Type[BaseTestImage] = BaseTestImage
    _TestImageDataFrameClass: Type[DFType] = DFType

    name: str
    """The unique name of this test case."""

    version: int
    """
    The version of this test case. A test case's version is automatically incremented whenever it is edited via
    [`TestCase.edit`][kolena.detection._internal.BaseTestCase.edit].
    """

    description: str
    """Free-form, human-readable description of this test case. Can be edited at any time via
    [`TestCase.edit`][kolena.detection._internal.BaseTestCase.edit].
    """

    _id: int
    _workflow: WorkflowType

    def __init__(
        self,
        name: str,
        workflow: WorkflowType,
        version: Optional[int] = None,
        description: Optional[str] = None,
        images: Optional[List[_TestImageClass]] = None,
        reset: bool = False,
    ):
        validate_name(name, FieldName.TEST_CASE_NAME)
        try:
            self._populate_from_other(self.load(name, version))
            if description is not None and self.description != description and not reset:
                log.warn("test case already exists, not updating description when reset=False")
            if images is not None:
                if self.version > 0 and not reset:
                    log.warn("not updating images for test case that has already been edited when reset=False")
                else:
                    self._hydrate(images, description)
        except NotFoundError:
            if version is not None:
                log.warn(f"creating new test case '{name}', ignoring provided version")
            self._populate_from_other(self._create(workflow, name, description, images))
        self._freeze()

    def _populate_from_other(self, other: "BaseTestCase") -> None:
        with self._unfrozen():
            self._id = other._id
            self.name = other.name
            self.version = other.version
            self.description = other.description
            self._workflow = other._workflow

    @classmethod
    def _from(cls, response: CoreAPI.EntityData) -> "BaseTestCase":
        obj = cls.__new__(cls)
        obj.name = response.name
        obj.version = response.version
        obj.description = response.description
        obj._id = response.id
        obj._workflow = WorkflowType(response.workflow)
        obj._freeze()
        return obj

    @classmethod
    def _create(
        cls,
        workflow: WorkflowType,
        name: str,
        description: Optional[str] = None,
        images: Optional[List[_TestImageClass]] = None,
    ) -> "BaseTestCase":
        """Create a new test case with the provided name."""
        validate_name(name, FieldName.TEST_CASE_NAME)
        request = CoreAPI.CreateRequest(name=name, description=description or "", workflow=workflow.value)
        res = krequests.post(endpoint_path=API.Path.CREATE.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        data = from_dict(data_class=CoreAPI.EntityData, data=res.json())
        obj = cls._create_from_data(data)
        log.info(f"created test case '{name}' (v{obj.version})")
        if images is not None:
            obj._hydrate(images)
        return obj

    @classmethod
    @validate_arguments(config=ValidatorConfig)
    def _load_by_name(cls, name: str, version: Optional[int] = None) -> CoreAPI.EntityData:
        """Load an existing test case with the provided name."""
        request = CoreAPI.LoadByNameRequest(name=name, version=version)
        res = krequests.put(endpoint_path=API.Path.LOAD_BY_NAME.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        obj = from_dict(data_class=CoreAPI.EntityData, data=res.json())
        log.info(f"loaded test case '{name}' (v{obj.version})")
        return obj

    @validate_arguments(config=ValidatorConfig)
    def _hydrate(self, images: List[_TestImageClass], description: Optional[str] = None) -> None:
        if len(images) == 0:
            log.warn("no images provided, unable to populate test case")
            return
        with self.edit(reset=True) as editor:
            if description is not None:
                editor.description(description)
            for image in images:
                editor.add(image)

    @classmethod
    def _create_from_data(cls, data: CoreAPI.EntityData) -> "BaseTestCase":
        assert_workflows_match(cls._workflow, data.workflow)
        return cls._from(data)

    @validate_arguments(config=ValidatorConfig)
    def load_images(self) -> List[_TestImageClass]:
        """Load all test images with their associated ground truths in this test case."""
        return list(self.iter_images())

    @validate_arguments(config=ValidatorConfig)
    def iter_images(self) -> Iterator[_TestImageClass]:
        """Iterate through all images with their associated ground truths in this test case."""
        log.info(f"loading test images for test case '{self.name}' (v{self.version})")
        init_request = CoreAPI.InitLoadContentsRequest(batch_size=BatchSize.LOAD_SAMPLES.value, test_case_id=self._id)
        for df in _BatchedLoader.iter_data(
            init_request=init_request,
            endpoint_path=API.Path.INIT_LOAD_IMAGES.value,
            df_class=self._TestImageDataFrameClass,
        ):
            for record in df.itertuples():
                yield self._TestImageClass._from_record(record)
        log.info(f"loaded test images for test case '{self.name}' (v{self.version})")

    @classmethod
    def create(
        cls,
        name: str,
        description: Optional[str] = None,
        images: Optional[List[_TestImageClass]] = None,
    ) -> "BaseTestCase":
        """
        Create a new test case with the provided name.

        :param name: The name of the new test case to create.
        :param description: Optional free-form description of the test case to create.
        :param images: Optionally specify a set of images to populate the test case.
        :return: The newly created test case.
        """
        return cls._create(cls._workflow, name, description, images)

    @classmethod
    def load(cls, name: str, version: Optional[int] = None) -> "BaseTestCase":
        """
        Load an existing test case with the provided name.

        :param name: The name of the test case to load.
        :param version: Optionally specify a particular version of the test case to load. Defaults to the latest version
            when unset.
        :return: The loaded test case.
        """
        data = cls._load_by_name(name, version)
        return cls._create_from_data(data)

    class Editor:
        """
        Interface to edit a test case. Create with [`TestCase.edit`][kolena.detection._internal.BaseTestCase.edit].
        """

        _TestImageClass: Type[BaseTestImage] = BaseTestImage
        _images: Dict[str, BaseTestImage]
        _reset: bool
        _description: str
        _initial_description: str

        def __init__(self, description: str, reset: bool) -> None:
            self._edited = False
            self._reset = reset
            self._description = description
            self._initial_description = description
            self._images: Dict[str, BaseTestImage] = OrderedDict()

        @validate_arguments(config=ValidatorConfig)
        def description(self, description: str) -> None:
            """
            Update the description of this test case.

            :param description: The new test case description.
            """
            if self._description == description:
                return
            self._description = description
            self._edited = True

        @validate_arguments(config=ValidatorConfig)
        def add(self, image: _TestImageClass) -> None:
            """
            Add a test image to the test case, targeting the `ground_truths` held by the image. When the test image
            already exists in the test case, its ground truth is overwritten.

            To filter the ground truths associated with a test image, see
            [`TestImage.filter`][kolena.detection.TestImage.filter].

            :param image: The test image to add to the test case, holding corresponding ground truths.
            """
            if image == self._images.get(image.locator, None):
                log.info(f"no op: {image.locator} already in test case")
                return
            self._images[image.locator] = image
            self._edited = True

        @validate_arguments(config=ValidatorConfig)
        def remove(self, image: _TestImageClass) -> None:
            """
            Remove the image from the test case.

            :param image: The image to remove.
            :raises KeyError: The image is not in the test case.
            """
            if image.locator not in self._images.keys():
                raise KeyError(f"unrecognized image: '{image.locator}' not in test case")
            self._images.pop(image.locator)
            self._edited = True

    @classmethod
    def _to_data_frame(cls, images: List[_TestImageClass]) -> _TestImageDataFrameClass:
        records = [cls._TestImageClass._to_record(image) for image in images]
        columns = cls._TestImageClass._meta_keys()
        df = pd.DataFrame(records, columns=columns)
        return cls._TestImageDataFrameClass(
            validate_df_schema(df, cls._TestImageDataFrameClass.get_schema(), trusted=True),
        )

    @contextmanager
    def edit(self, reset: bool = False) -> Iterator[Editor]:
        """
        Edit this test case in a context:

        ```python
        with test_case.edit() as editor:
            # perform as many editing actions as desired
            editor.add(...)
            editor.remove(...)
        ```

        Changes are committed to Kolena when the context is exited.

        :param reset: Clear all existing test samples in the test case.
        """
        editor = self.Editor(self.description, reset)
        editor._TestImageClass = self._TestImageClass
        if not reset:
            for image in self.iter_images():
                editor.add(image)
            editor._edited = False

        yield editor

        # no-op contexts have no effect, do not bump version
        if not editor._edited:
            return

        log.info(f"editing test case '{self.name}' (v{self.version})")
        init_response = init_upload()
        df = self._to_data_frame(list(editor._images.values()))
        df_serialized = df.as_serializable()
        upload_data_frame(df=df_serialized, batch_size=BatchSize.UPLOAD_RECORDS.value, load_uuid=init_response.uuid)

        request = CoreAPI.CompleteEditRequest(
            test_case_id=self._id,
            current_version=self.version,
            description=editor._description,
            reset=editor._reset,
            uuid=init_response.uuid,
        )
        complete_res = krequests.put(
            endpoint_path=API.Path.COMPLETE_EDIT.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(complete_res)
        test_case_data = from_dict(data_class=CoreAPI.EntityData, data=complete_res.json())
        self._populate_from_other(self._create_from_data(test_case_data))
        log.success(f"edited test case '{self.name}' (v{self.version})")



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

from deprecation import deprecated
from pydantic import validate_arguments

from kolena._api.v1.detection import TestImage as API
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.consts import BatchSize
from kolena._utils.validators import ValidatorConfig
from kolena.detection._datatypes import TestImageDataFrame
from kolena.detection._internal import BaseTestImage
from kolena.detection._internal.metadata import _from_dict
from kolena.detection._internal.metadata import _to_dict
from kolena.detection._internal.metadata import MetadataElement
from kolena.detection.ground_truth import GroundTruth


class TestImage(BaseTestImage):
    """
    Test image comprising a single image locator.
    """

    locator: str
    """Bucket locator for the provided test sample, e.g. `gs://my-bucket/path/to/image.png`."""

    dataset: str
    """Dataset this test image belongs to. Empty when unspecified."""

    metadata: Dict[str, MetadataElement]
    """Metadata associated with this test image. Surfaced during test runs."""

    ground_truths: List[GroundTruth]
    """List of [`GroundTruth`][kolena.detection.ground_truth.GroundTruth] annotations associated with this image."""

    @validate_arguments(config=ValidatorConfig)
    def __init__(
        self,
        locator: str,
        dataset: Optional[str] = None,
        ground_truths: Optional[List[GroundTruth]] = None,
        metadata: Dict[str, MetadataElement] = None,
    ):
        super().__init__(locator, dataset, metadata)
        self.ground_truths = ground_truths or []

    def filter(self, predicate: Callable[[GroundTruth], bool]) -> "TestImage":
        """
        Return a copy of this test image with ground truths filtered to only those that match the provided predicate.

        :param predicate: Function accepting a [`GroundTruth`][kolena.detection.ground_truth.GroundTruth] and returning
            a boolean indicating whether or not to include the ground truth.
        :return: A new test image with ground truths filtered by the predicate.
        """
        return TestImage(**{**self._fields(), "ground_truths": [gt for gt in self.ground_truths if predicate(gt)]})

    @classmethod
    def _meta_keys(cls) -> List[str]:
        return ["locator", "dataset", "ground_truths", "metadata"]

    @classmethod
    def _from_record(cls, record: Any) -> "TestImage":
        return TestImage(
            record.locator,
            dataset=record.dataset,
            ground_truths=[GroundTruth._from_dict(gt_dict) for gt_dict in record.ground_truths or []],
            metadata=_from_dict(record.metadata),
        )

    @classmethod
    def _to_record(cls, image: "TestImage") -> Tuple[str, Optional[str], List[Dict[str, Any]], Dict[str, Any]]:
        return (image.locator, image.dataset, [gt._to_dict() for gt in image.ground_truths], _to_dict(image.metadata))

    # TODO: remove implementation in favor of Frozen.__eq__ once ground_truth ordering is ensured upstream
    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and {
            **self.__dict__,
            "ground_truths": self._sort_ground_truths(self.ground_truths),
        } == {**other.__dict__, "ground_truths": self._sort_ground_truths(other.ground_truths)}

    @staticmethod
    def _sort_ground_truths(ground_truths: List[GroundTruth]) -> List[GroundTruth]:
        return sorted(ground_truths, key=lambda gt: json.dumps(gt._to_dict(), sort_keys=True))


@deprecated(details="use `TestCase.load_images`", deprecated_in="0.26.0")
@validate_arguments(config=ValidatorConfig)
def load_images(dataset: Optional[str] = None) -> List[TestImage]:
    """
    !!! warning "Deprecated: since `0.26.0`"
        Please use [`TestCase.load_images`][kolena.detection._internal.test_case.BaseTestCase.load_images] instead.

    Load a list of [`TestImage`][kolena.detection.TestImage] samples registered in Kolena.

    :param dataset: Optionally specify the single dataset to be retrieved. By default, images from all
        datasets are returned.
    """
    return list(iter_images(dataset))


@deprecated(details="use `TestCase.iter_images`", deprecated_in="0.26.0")
@validate_arguments(config=ValidatorConfig)
def iter_images(dataset: Optional[str] = None) -> Iterator[TestImage]:
    """
    !!! warning "Deprecated: since `0.26.0`"
        Please use [`TestCase.iter_images`][kolena.detection._internal.test_case.BaseTestCase.iter_images] instead.

    Return iterator over [`TestImage`][kolena.detection.TestImage] samples registered in Kolena. Images are lazily
    loaded in chunks to facilitate working with large datasets that are cumbersome to hold in memory.

    :param dataset: Optionally specify the single dataset to be retrieved. By default, images from all
        datasets are returned.
    """
    init_request = API.InitLoadImagesRequest(dataset=dataset, batch_size=BatchSize.LOAD_RECORDS.value)
    for df in _BatchedLoader.iter_data(
        init_request=init_request,
        endpoint_path=API.Path.INIT_LOAD_IMAGES.value,
        df_class=TestImageDataFrame,
    ):
        for record in df.itertuples():
            yield TestImage._from_record(record)



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noreorder
from .datatypes import EmbeddingDataFrame
from .datatypes import ImageDataFrame
from .datatypes import ImageResultDataFrame
from .datatypes import PairDataFrame
from .datatypes import PairResultDataFrame
from .datatypes import TestImageDataFrame
from .datatypes import TestCaseDataFrame
from .test_case import TestCase
from .test_suite import TestSuite
from .test_images import TestImages
from .model import Model
from .model import InferenceModel
from .test_run import TestRun
from .test_run import test

__all__ = [
    "EmbeddingDataFrame",
    "ImageDataFrame",
    "ImageResultDataFrame",
    "PairDataFrame",
    "PairResultDataFrame",
    "TestImageDataFrame",
    "TestCaseDataFrame",
    "TestCase",
    "TestSuite",
    "TestImages",
    "Model",
    "InferenceModel",
    "TestRun",
    "test",
]



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Dict
from urllib.parse import urlencode
from urllib.parse import urlparse

from kolena._utils.state import _ClientState
from kolena._utils.state import get_client_state
from kolena._utils.state import get_endpoint_with_baseurl
from kolena.errors import InvalidClientStateError


def get_endpoint(endpoint_path: str) -> str:
    client_state = get_client_state()
    if client_state.base_url is None:
        raise InvalidClientStateError("missing base_url")
    return get_endpoint_with_baseurl(client_state.base_url, endpoint_path)


def get_platform_url() -> str:
    return _get_platform_url(get_client_state())


# pure function for testing
def _get_platform_url(client_state: _ClientState) -> str:
    return f"{_get_platform_origin(client_state)}/{client_state.tenant}"


def _get_platform_origin(client_state: _ClientState) -> str:
    if client_state.base_url is None:
        raise InvalidClientStateError("missing base_url")
    base_url = urlparse(client_state.base_url)
    if base_url.hostname == "localhost":
        return "http://localhost:3000"
    gateway_subdomain, *_ = base_url.hostname.split(".")
    subdomain = "trunk" if "trunk" in gateway_subdomain else "app"
    return f"https://{subdomain}.kolena.io"


def get_results_url(workflow: str, model_id: int, test_suite_id: int) -> str:
    return _get_results_url(get_client_state(), workflow, model_id, test_suite_id)


# pure function for testing
def _get_results_url(client_state: _ClientState, workflow: str, model_id: int, test_suite_id: int) -> str:
    from kolena._api.v1.workflow import WorkflowType  # deferred import to avoid circular dependencies

    platform_url = _get_platform_url(client_state)
    params: Dict[str, Any] = dict(modelIds=model_id, testSuiteId=test_suite_id)
    if workflow == WorkflowType.FR.value:
        path = "results/fr"
    elif workflow == WorkflowType.DETECTION.value:
        path = "results/object-detection"
    elif workflow == WorkflowType.CLASSIFICATION.value:
        path = "results/classification"
    else:
        path = "results"
        params["workflow"] = workflow
    return f"{platform_url}/{path}?{urlencode(params)}"


def get_test_suite_url(test_suite_id: int) -> str:
    return _get_test_suite_url(get_client_state(), test_suite_id)


def _get_test_suite_url(client_state: _ClientState, test_suite_id: int) -> str:
    platform_url = _get_platform_url(client_state)
    return f"{platform_url}/testing?{urlencode(dict(testSuiteId=test_suite_id))}"


def get_model_url(model_id: int) -> str:
    return _get_model_url(get_client_state(), model_id)


def _get_model_url(client_state: _ClientState, model_id: int) -> str:
    platform_url = _get_platform_url(client_state)
    return f"{platform_url}/models?{urlencode(dict(modelIds=model_id))}"



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

from pydantic import validate_arguments

from kolena._utils import log
from kolena._utils.validators import ValidatorConfig
from kolena.classification import InferenceModel
from kolena.classification import Model
from kolena.classification import TestConfig
from kolena.classification import TestImage
from kolena.classification import TestSuite
from kolena.classification.test_config import AccuracyOptimal
from kolena.detection._datatypes import ImageResultDataFrame
from kolena.detection._internal import BaseTestRun
from kolena.detection._internal.datatypes import LoadTestImagesDataFrame
from kolena.detection._internal.metadata import _from_dict
from kolena.detection._internal.test_run import CustomMetricsCallback
from kolena.detection.inference import ClassificationLabel


class TestRun(BaseTestRun):
    """
    Interface to run tests for a [`Model`][kolena.classification.Model] on a
    [`TestSuite`][kolena.classification.TestSuite]. Any in-progress tests for this model on these suites are resumed.

    For a streamlined interface, see [`test`][kolena.classification.test].

    :param model: The model being tested.
    :param test_suite: The test suite on which to test the model.
    :param test_config: Optionally specify a configuration, e.g.
        [`FixedGlobalThreshold`][kolena.classification.FixedGlobalThreshold], to customize the metrics evaluation logic
        for this test run. Defaults to [`AccuracyOptimal`][kolena.classification.test_config.AccuracyOptimal] if
        unspecified.
    :param custom_metrics_callback: Optionally specify a callback function to compute custom metrics for each test-case.
        The callback would be passed inferences of images in each testcase and should return a dictionary with metric
        name as key and metric value as value.
    :param reset: Overwrites existing inferences if set.
    """

    _TestImageClass = TestImage
    _InferenceClass = Tuple[str, float]
    _ImageResultDataFrameClass: Type[ImageResultDataFrame] = ImageResultDataFrame
    _LoadTestImagesDataFrameClass: Type[LoadTestImagesDataFrame] = LoadTestImagesDataFrame

    @validate_arguments(config=ValidatorConfig)
    def __init__(
        self,
        model: Model,
        test_suite: TestSuite,
        test_config: Optional[TestConfig] = None,
        custom_metrics_callback: Optional[CustomMetricsCallback[_TestImageClass, _InferenceClass]] = None,
        reset: bool = False,
    ):
        config = AccuracyOptimal() if test_config is None else test_config
        super().__init__(
            model,
            test_suite,
            config=config._to_run_config(),
            custom_metrics_callback=custom_metrics_callback,
            reset=reset,
        )

    @validate_arguments(config=ValidatorConfig)
    def add_inferences(self, image: _TestImageClass, inferences: Optional[List[_InferenceClass]]) -> None:
        super().add_inferences(
            image,
            [ClassificationLabel(label, confidence) for label, confidence in inferences]
            if inferences is not None
            else None,
        )

    def _image_from_load_image_record(self, record: Any) -> _TestImageClass:
        self._locator_to_image_id[record.locator] = record.test_sample_id
        return self._TestImageClass(
            locator=record.locator,
            dataset=record.dataset,
            metadata=_from_dict(record.metadata),
            labels=[],  # note that no ground truth labels are loaded during testing
        )


@validate_arguments(config=ValidatorConfig)
def test(
    model: InferenceModel,
    test_suite: TestSuite,
    test_config: Optional[TestConfig] = None,
    custom_metrics_callback: Optional[CustomMetricsCallback[TestImage, Tuple[str, float]]] = None,
    reset: bool = False,
) -> None:
    """
    Test the provided [`InferenceModel`][kolena.classification.InferenceModel] on a
    [`TestSuite`][kolena.classification.TestSuite]. Any tests already in progress for this model on these suites are
    resumed.

    :param model: The model being tested, complete with an
        [`InferenceModel.infer`][kolena.classification.InferenceModel.infer] function to perform inference.
    :param test_suite: The test suite on which to test the model.
    :param test_config: Optionally specify a configuration, e.g.
        [`FixedGlobalThreshold`][kolena.classification.FixedGlobalThreshold], to customize the metrics evaluation logic
        for this test run. Defaults to [`AccuracyOptimal`][kolena.classification.test_config.AccuracyOptimal] if
        unspecified.
    :param custom_metrics_callback: Optionally specify a callback function to compute custom metrics for each test case.
        The callback would be passed inferences of images in each testcase and should return a dictionary with metric
        name as key and metric value as value.
    :param reset: Overwrites existing inferences if set.
    """
    with TestRun(
        model,
        test_suite,
        test_config=test_config,
        custom_metrics_callback=custom_metrics_callback,
        reset=reset,
    ) as test_run:
        log.info("performing inference")
        for image in log.progress_bar(test_run.iter_images()):
            test_run.add_inferences(image, model.infer(image))
        log.success("performed inference")



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

from pydantic import validate_arguments

from kolena._utils.geometry import validate_polygon
from kolena._utils.inference_validators import validate_confidence
from kolena._utils.inference_validators import validate_label
from kolena._utils.validators import ValidatorConfig
from kolena.detection._internal import Inference as _Inference
from kolena.detection._internal import InferenceType


class Inference(_Inference):
    """
    Base class for model inferences associated with an image.

    See concrete implementations [`BoundingBox`][kolena.detection.inference.BoundingBox], and
    [`SegmentationMask`][kolena.detection.inference.SegmentationMask] for details.
    """

    @classmethod
    def _from_dict(cls, ground_truth: Dict[str, Any]) -> "Inference":
        data_type = ground_truth["data_type"]
        data_object = ground_truth["data_object"]
        if data_type == InferenceType.CLASSIFICATION_LABEL.value:
            return ClassificationLabel(label=data_object["label"], confidence=data_object["confidence"])
        if data_type == InferenceType.BOUNDING_BOX.value:
            top_left, bottom_right = data_object["points"]
            return BoundingBox(
                label=data_object["label"],
                confidence=data_object["confidence"],
                top_left=top_left,
                bottom_right=bottom_right,
            )
        if data_type == InferenceType.SEGMENTATION_MASK.value:
            return SegmentationMask(
                label=data_object["label"],
                confidence=data_object["confidence"],
                points=data_object["points"],
            )
        raise ValueError(f"invalid dictionary provided, unrecognized data type '{data_type}'")


# NOTE: intentionally leave undocumented; not a part of the public API
class ClassificationLabel(Inference):
    label: str
    confidence: float

    @validate_arguments(config=ValidatorConfig)
    def __init__(self, label: str, confidence: float):
        validate_label(label)
        validate_confidence(confidence)
        self.label = label
        self.confidence = confidence

    def _to_dict(self) -> Dict[str, Any]:
        return dict(
            data_type=InferenceType.CLASSIFICATION_LABEL.value,
            data_object=dict(label=self.label, confidence=self.confidence),
        )


class BoundingBox(Inference):
    """
    Inference representing a bounding box.

    Point coordinates should be in `(x, y)` format, as absolute pixel values.
    """

    label: str
    """Label to associate with this bounding box."""

    confidence: float
    """Confidence score associated with this inference."""

    top_left: Tuple[float, float]
    """Point in `(x, y)` pixel coordinates representing the top left corner of the bounding box."""

    bottom_right: Tuple[float, float]
    """Point in `(x, y)` pixel coordinates representing the bottom right corner of the bounding box."""

    @validate_arguments(config=ValidatorConfig)
    def __init__(self, label: str, confidence: float, top_left: Tuple[float, float], bottom_right: Tuple[float, float]):
        validate_label(label)
        validate_confidence(confidence)
        self.label = label
        self.confidence = confidence
        self.top_left = top_left
        self.bottom_right = bottom_right

    def _to_dict(self) -> Dict[str, Any]:
        return dict(
            data_type=InferenceType.BOUNDING_BOX.value,
            data_object=dict(label=self.label, confidence=self.confidence, points=[self.top_left, self.bottom_right]),
        )


class SegmentationMask(Inference):
    """
    Inference data object representing a segmentation mask.

    Point coordinates should be in `(x, y)` format, as absolute pixel values.

    :raises ValueError: When fewer than three points are provided.
    """

    label: str
    """Label to associate with this segmentation mask."""

    confidence: float
    """Confidence score associated with this inference."""

    points: List[Tuple[float, float]]
    """
    Polygon corresponding to the vertices of the segmentation mask. Must have at least three distinct elements, and
    may not cross itself or touch itself at any point.
    """

    @validate_arguments(config=ValidatorConfig)
    def __init__(self, label: str, confidence: float, points: List[Tuple[float, float]]):
        validate_label(label)
        validate_confidence(confidence)
        validate_polygon(points)
        self.label = label
        self.confidence = confidence
        self.points = points

    def _to_dict(self) -> Dict[str, Any]:
        return dict(
            data_type=InferenceType.SEGMENTATION_MASK.value,
            data_object=dict(label=self.label, confidence=self.confidence, points=self.points),
        )



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
from abc import ABC
from abc import abstractmethod
from collections import OrderedDict
from types import TracebackType
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

import pandera as pa
from pydantic import validate_arguments

from kolena._api.v1.detection import CustomMetrics
from kolena._api.v1.detection import Metrics
from kolena._api.v1.detection import TestRun as API
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.batched_load import DFType
from kolena._utils.batched_load import init_upload
from kolena._utils.batched_load import upload_data_frame_chunk
from kolena._utils.consts import BatchSize
from kolena._utils.datatypes import LoadableDataFrame
from kolena._utils.frozen import Frozen
from kolena._utils.instrumentation import report_crash
from kolena._utils.instrumentation import WithTelemetry
from kolena._utils.serde import from_dict
from kolena._utils.validators import ValidatorConfig
from kolena.detection._internal import BaseModel
from kolena.detection._internal import BaseTestImage
from kolena.detection._internal import BaseTestSuite
from kolena.detection._internal import Inference
from kolena.detection._internal.model import SampleInferences
from kolena.errors import CustomMetricsException
from kolena.errors import IncorrectUsageError
from kolena.errors import InputValidationError
from kolena.errors import WorkflowMismatchError

_ImageDataFrame = Union[pa.typing.DataFrame, LoadableDataFrame]
InferenceType = TypeVar("InferenceType")

CustomMetricsCallback = Callable[[List[SampleInferences]], CustomMetrics]
"""Signature for a custom metrics computation function."""


class BaseTestRun(ABC, Frozen, WithTelemetry):
    """
    The base class for [`kolena.classification.TestRun`][kolena.classification.TestRun] and
    [`kolena.detection.TestRun`][kolena.detection.TestRun].
    """

    _TestImageClass: Type[BaseTestImage] = BaseTestImage
    _InferenceClass: Type[InferenceType] = InferenceType
    _ImageResultDataFrameClass: Type[_ImageDataFrame] = _ImageDataFrame
    _LoadTestImagesDataFrameClass: Type[DFType] = DFType

    @validate_arguments(config=ValidatorConfig)
    def __init__(
        self,
        model: BaseModel,
        test_suite: BaseTestSuite,
        config: Optional[Metrics.RunConfig] = None,
        custom_metrics_callback: Optional[CustomMetricsCallback[_TestImageClass, _InferenceClass]] = None,
        reset: bool = False,
    ):
        if model._workflow != test_suite._workflow:
            raise WorkflowMismatchError(
                f"mismatching test suite workflow for model of type '{model._workflow}': '{test_suite._workflow}'",
            )

        if reset:
            log.warn("overwriting existing inferences from this model (reset=True)")
        else:
            log.info("not overwriting any existing inferences from this model (reset=False)")

        request = API.CreateOrRetrieveRequest(
            model_id=model._id,
            test_suite_ids=[test_suite._id],
            config=config,
        )
        res = krequests.post(
            endpoint_path=API.Path.CREATE_OR_RETRIEVE.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(res)
        response = from_dict(data_class=API.CreateOrRetrieveResponse, data=res.json())
        self._id = response.test_run_id
        self._model = model
        self._test_suite = test_suite
        self._locator_to_image_id: Dict[str, int] = {}
        self._inferences: Dict[int, List[Optional[Inference]]] = OrderedDict()
        self._ignored_image_ids: List[int] = []
        self._upload_uuid: Optional[str] = None
        self._n_inferences = 0
        self._custom_metrics_callback: CustomMetricsCallback = custom_metrics_callback
        self._active = False
        self._reset = reset
        # note not calling self._freeze()

    def __enter__(self) -> "BaseTestRun":
        self._active = True
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self._upload_chunk()
        self._finalize_upload()
        self._submit_custom_metrics()
        self._active = False
        if exc_type is not None:
            report_crash(self._id, API.Path.MARK_CRASHED.value)

    @validate_arguments(config=ValidatorConfig)
    def add_inferences(self, image: _TestImageClass, inferences: Optional[List[_InferenceClass]]) -> None:
        """
        Add inferences for a test image to the test run results.

        :param image: The image that inferences are evaluated on.
        :param inferences: List of inferences corresponding to the image.
        """
        self._assert_active()

        if image.locator not in self._locator_to_image_id.keys():
            raise InputValidationError(
                f"Unrecognized locator '{image.locator}'. Images must be loaded and processed in the same context",
            )
        image_id = self._locator_to_image_id[image.locator]
        if inferences is None:
            self._ignored_image_ids.append(image_id)
            self._n_inferences += 1

        else:
            context_image_inferences = self._inferences.get(image_id, [])
            if len(inferences) == 0:
                context_image_inferences.append(None)
                self._n_inferences += 1
            else:
                context_image_inferences.extend(inferences)
                self._n_inferences += len(inferences)

            self._inferences[image_id] = context_image_inferences

        if self._n_inferences >= BatchSize.UPLOAD_RESULTS.value:
            self._upload_chunk()

    @validate_arguments(config=ValidatorConfig)
    def iter_images(self) -> Iterator[_TestImageClass]:
        """Returns an iterator of all remaining images that need inferences evaluated."""
        self._assert_active()
        for df_image_batch in self._iter_image_batch():
            for record in df_image_batch.itertuples():
                yield self._image_from_load_image_record(record)

    @validate_arguments(config=ValidatorConfig)
    def load_images(self, batch_size: int = BatchSize.LOAD_SAMPLES.value) -> List[_TestImageClass]:
        """
        Returns a list of images that still need inferences evaluated, bounded in count by `batch_size`. Note that image
        ground truths will be excluded from the returned batch of images.

        :param batch_size: The maximum number of images to retrieve.
        """
        self._assert_active()
        log.info("loading batch of images for test run")
        try:
            df_image_batch = next(self._iter_image_batch(batch_size=batch_size))
        except StopIteration:
            # no more images
            return []
        log.info("loaded batch of images for test run")
        return [self._image_from_load_image_record(record) for record in df_image_batch.itertuples()]

    @validate_arguments(config=ValidatorConfig)
    def _iter_image_batch(
        self,
        batch_size: int = BatchSize.LOAD_SAMPLES.value,
    ) -> Iterator[_LoadTestImagesDataFrameClass]:
        if batch_size <= 0:
            raise InputValidationError(f"invalid batch_size '{batch_size}': expected positive integer")
        init_request = API.InitLoadRemainingImagesRequest(
            test_run_id=self._id,
            batch_size=batch_size,
            load_all=self._reset,
        )
        yield from _BatchedLoader.iter_data(
            init_request=init_request,
            endpoint_path=API.Path.INIT_LOAD_REMAINING_IMAGES.value,
            df_class=self._LoadTestImagesDataFrameClass,
        )

    @validate_arguments(config=ValidatorConfig)
    def _upload_chunk(self) -> None:
        if self._n_inferences == 0:
            # Bail if this happens to being run by fencepost immediately after being run by add_inference
            return

        log.info(f"uploading {self._n_inferences} inferences for test run")
        if self._upload_uuid is None:
            init_response = init_upload()
            self._upload_uuid = init_response.uuid

        df_chunk = self._ImageResultDataFrameClass.from_image_inference_mapping(
            self._id,
            self._model._id,
            self._inferences,
            self._ignored_image_ids,
        )
        df_chunk_serializable = df_chunk.as_serializable()
        upload_data_frame_chunk(df_chunk_serializable, load_uuid=self._upload_uuid)
        log.success(f"uploaded {self._n_inferences} inferences for test run")
        self._n_inferences = 0
        self._inferences = OrderedDict()

    def _finalize_upload(self) -> None:
        if self._upload_uuid is None:
            # nothing was uploaded
            return

        log.info("finalizing inference upload for test run")
        request = API.UploadImageResultsRequest(uuid=self._upload_uuid, test_run_id=self._id, reset=self._reset)
        finalize_res = krequests.put(
            endpoint_path=API.Path.UPLOAD_IMAGE_RESULTS.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(finalize_res)
        log.success("finalized inference upload for test run")

    def _assert_active(self) -> None:
        if not self._active:
            raise IncorrectUsageError("test run must be used inside a context manager")

    @abstractmethod
    def _image_from_load_image_record(self, record: Any) -> _TestImageClass:
        ...

    @validate_arguments(config=ValidatorConfig)
    def _compute_custom_metrics(self) -> Dict[int, Dict[int, CustomMetrics]]:
        log.info("computing custom metrics for test run")
        test_case_metrics = {}
        custom_metrics = {}  # { test_suite_id: { test_case_id: CustomMetrics } }

        test_suite = self._test_suite
        log.info(f"computing custom metrics for test suite '{test_suite.name}'")
        test_suite_metrics = {}
        all_inferences = self._model.load_inferences_by_test_case(test_suite)
        for test_case in log.progress_bar(test_suite.test_cases):
            test_case_id = test_case._id
            if test_case_id not in test_case_metrics:
                inferences = all_inferences.get(test_case_id, [])
                try:
                    test_case_metrics[test_case_id] = self._custom_metrics_callback(inferences)
                except Exception as e:
                    raise CustomMetricsException(
                        f"Error encountered computing custom metrics for test case '{test_case.name}'",
                    ) from e

            test_suite_metrics[test_case_id] = test_case_metrics[test_case_id]

        custom_metrics[test_suite._id] = test_suite_metrics
        log.success(f"computed custom metrics for test suite '{test_suite.name}'")

        log.success("computed custom metrics for test run")
        return custom_metrics

    def _submit_custom_metrics(self) -> None:
        if self._custom_metrics_callback is None:
            return

        log.info("computing and uploading custom metrics for test run")
        custom_metrics = self._compute_custom_metrics()
        request = API.UpdateCustomMetricsRequest(model_id=self._model._id, metrics=custom_metrics)
        res = krequests.put(
            endpoint_path=API.Path.UPLOAD_CUSTOM_METRICS.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(res)
        log.success("computed and uploaded custom metrics for test run")
